package main

import (
	"crypto/sha256"
	"encoding/hex"
	"encoding/json"
	"flag"
	"fmt"
	"io"
	"io/ioutil"
	"log"
	"os"
	"path/filepath"
	"runtime"
	"sync"
)

// Fdhashes struct for holding all informations about one folder.
type Fdhashes struct {
	Hashes map[string]string
}

var hashes map[string]Fdhashes

var wg sync.WaitGroup
var mu sync.RWMutex
var driveLetter string

var rewrite bool

func init() {
	flag.BoolVar(&rewrite, "rewrite", false, "rewrite all fhhashes files.")
}

func main() {
	log.Println("starting GoHash")
	runtime.GOMAXPROCS(5)
	hashes = make(map[string]Fdhashes)
	flag.Parse()
	myFile := flag.Arg(0)
	file, err := os.Stat(myFile)
	if os.IsNotExist(err) {
		log.Fatalln("File does not exists:", myFile)
	}
	if file.IsDir() {
		log.Println("start with folder:", myFile)
		driveLetter = ""
		if runtime.GOOS == "windows" {
			driveLetter = filepath.VolumeName(myFile) + "/"
		}
		processFolder(myFile)
	} else {
		log.Printf("file %s has hash %s\n", myFile, getSha256Hash(myFile))
	}

	fmt.Println("waiting")
	wg.Wait()
	log.Println("done")
}

func getSha256Hash(fileStr string) string {
	f, err := os.Open(fileStr)
	if err != nil {
		log.Fatal(err)
	}
	defer f.Close()

	h := sha256.New()
	if _, err := io.Copy(h, f); err != nil {
		log.Fatal(err)
	}
	return hex.EncodeToString(h.Sum(nil))
}

func outputHash(fileStr string) {
	var hashFile Fdhashes
	doHash := true
	defer wg.Done()
	dir, fileName := filepath.Split(fileStr)
	if fileName == ".fdhashes3" {
		return
	}
	mu.Lock()
	hashFile, ok := hashes[dir]
	if !ok {
		_, err := os.Stat(dir + ".fdhashes3")
		if os.IsNotExist(err) {
			hashFile = Fdhashes{Hashes: make(map[string]string)}
		} else {
			hashFile = loadHashfile(dir + ".fdhashes3")
		}
		hashes[dir] = hashFile
		saveHashfile(hashFile, dir)
	}

	_, ok = hashFile.Hashes[fileName]
	doHash = !ok
	hashFile = hashes[dir]
	mu.Unlock()
	if doHash {
		hash := getSha256Hash(fileStr)
		mu.Lock()
		hashFile.Hashes[fileName] = hash
		saveHashfile(hashFile, dir)
		mu.Unlock()
		log.Printf("file \"%s\" has hash \"%s\"\n", fileStr, hash)
	}
}

var count int

func processFolder(folder string) {
	count = 0
	err := filepath.Walk(folder, func(path string, info os.FileInfo, err error) error {
		count++
		if (count % 100) == 0 {
			fmt.Print(".")
		}
		if (count % 10000) == 0 {
			fmt.Println()
		}
		filename := info.Name()
		if filename[0:1] != "." {
			if info.IsDir() {
				fmt.Println(path)
			}
			if !info.IsDir() {
				wg.Add(1)
				go outputHash(path)
			}
		}
		return nil
	})
	if err != nil {
		panic(err)
	}
}

func saveHashfile(hashFile Fdhashes, dir string) {
	b, err := json.Marshal(hashFile)
	if err != nil {
		fmt.Println(err)
		return
	}
	err = ioutil.WriteFile(dir+".fdhashes3", b, 0644)
	if err != nil {
		panic(err)
	}
}

func loadHashfile(filename string) Fdhashes {
	data := Fdhashes{Hashes: make(map[string]string)}
	if !rewrite {
		file, err := ioutil.ReadFile(filename)
		if err != nil {
			panic(err)
		}
		err = json.Unmarshal([]byte(file), &data)
		if err != nil {
			panic(err)
		}
	}
	data := Fdhashes{}
	err = json.Unmarshal([]byte(file), &data)
	if err != nil {
		log.Fatalf("can't read file %s", filename)
	}
	return data
}
